import errno
from io import BytesIO
import os
import time

try:
    import PIL
except ImportError:
    # Pillow / PIL is optional
    PIL = None
else:
    from PIL import Image

from flask import Response, current_app, render_template, stream_with_context
from flask.views import MethodView
from werkzeug.exceptions import NotFound, Forbidden

from ..constants import COMPLETE, FILENAME, LOCKED, SIZE, TIMESTAMP_DOWNLOAD, TYPE
from ..utils.date_funcs import delete_if_lifetime_over
from ..utils.permissions import ADMIN, READ, may


class DownloadView(MethodView):
    content_disposition = 'attachment'  # to trigger download

    def err_incomplete(self, item, error):
        return render_template('error.html', heading=item.meta[FILENAME], body=error), 409

    def stream(self, item, start, limit):
        with item as _item:
            # Stream content from storage
            offset = max(0, start)
            while offset < limit:
                buf = _item.data.read(min(limit - offset, 16 * 1024), offset)
                offset += len(buf)
                yield buf
            item.meta[TIMESTAMP_DOWNLOAD] = int(time.time())

    def response(self, item, name):
        ct = item.meta[TYPE]
        dispo = self.content_disposition
        if dispo != 'attachment':
            # no simple download, so we must be careful about XSS
            if ct.startswith("text/"):
                ct = 'text/plain'  # only send simple plain text

        ret = Response(stream_with_context(self.stream(item, 0, item.data.size)))
        ret.headers['Content-Disposition'] = '{}; filename="{}"'.format(
            dispo, item.meta[FILENAME])
        ret.headers['Content-Length'] = item.meta[SIZE]
        ret.headers['Content-Type'] = ct
        ret.headers['X-Content-Type-Options'] = 'nosniff'  # yes, we really mean it
        return ret

    def get(self, name):
        if not may(READ):
            raise Forbidden()
        try:
            item = current_app.storage.openwrite(name)
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise NotFound()
            raise

        try:
            need_close = True
            if not item.meta[COMPLETE]:
                return self.err_incomplete(item, 'Upload incomplete. Try again later.')

            if item.meta[LOCKED] and not may(ADMIN):
                raise Forbidden()

            if delete_if_lifetime_over(item, name):
                raise NotFound()
            need_close = False
        finally:
            if need_close:
                item.close()

        return self.response(item, name)


class InlineView(DownloadView):
    content_disposition = 'inline'  # to trigger viewing in browser, for some types


class ThumbnailView(InlineView):
    thumbnail_size = 192, 108
    thumbnail_data = """\
        <?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <svg width="108" height="108" viewBox="0 0 108 108" xmlns="http://www.w3.org/2000/svg">
        <rect x="1" y="1" width="106" height="106" fill="whitesmoke" stroke-width="2" stroke="blue" />
            <line x1="1" y1="1" x2="106" y2="106" stroke="blue" stroke-width="2" />
            <line x1="1" y1="106" x2="106" y2="0" stroke="blue" stroke-width="2" />
        </svg>""".strip().encode()

    def err_incomplete(self, item, error):
        return b'', 409  # conflict

    def response(self, item, name):
        sz = item.meta[SIZE]
        fn = item.meta[FILENAME]
        ct = item.meta[TYPE]
        unsupported = PIL is None or ct not in {'image/jpeg', 'image/png', 'image/gif', 'image/webp'}
        if unsupported:
            # return a placeholder thumbnail for unsupported item types
            ret = Response(self.thumbnail_data)
            ret.headers['Content-Length'] = len(self.thumbnail_data)
            ret.headers['Content-Type'] = 'image/svg+xml'
            ret.headers['X-Content-Type-Options'] = 'nosniff'  # yes, we really mean it
            return ret

        if ct in ('image/jpeg', ):
            thumbnail_type = 'jpeg'
        elif ct in ('image/png', 'image/gif'):
            thumbnail_type = 'png'
        elif ct in ('image/webp', ):
            thumbnail_type = 'webp'
        else:
            raise ValueError('unrecognized image content type')

        # compute thumbnail data "on the fly"
        with BytesIO(item.data.read(sz, 0)) as img_bio, BytesIO() as thumbnail_bio:
            with Image.open(img_bio) as img:
                img.thumbnail(self.thumbnail_size)
                img.save(thumbnail_bio, thumbnail_type)
            thumbnail_data = thumbnail_bio.getvalue()

        name, ext = os.path.splitext(fn)
        thumbnail_fn = '{}-thumb.{}'.format(name, thumbnail_type)

        ret = Response(thumbnail_data)
        ret.headers['Content-Disposition'] = '{}; filename="{}"'.format(
            self.content_disposition, thumbnail_fn)
        ret.headers['Content-Length'] = len(thumbnail_data)
        ret.headers['Content-Type'] = 'image/%s' % thumbnail_type
        ret.headers['X-Content-Type-Options'] = 'nosniff'  # yes, we really mean it
        return ret



from pygments.formatters.html import HtmlFormatter


class CustomHtmlFormatter(HtmlFormatter):
    """Custom HTML formatter. Adds an option to wrap lines into HTML <p> tags."""
    def __init__(self, **options):
        super().__init__(**options)
        self.lineparagraphs = options.get('lineparagraphs', '')

    def _wrap_lineparagraphs(self, inner):
        """Wrap lines into <p> tags

        :param inner: iterator of tuples of format (code, text)
        :return: iterator of tuples containing updated wrapped lines
        """
        s = self.lineparagraphs
        i = self.linenostart - 1
        for t, line in inner:
            if t:
                i += 1
                yield 1, '<p id="%s-%d">%s</p>' % (s, i, line)
            else:
                yield 0, line

    def format_unencoded(self, tokensource, outfile):
        """
        The formatting process uses several nested generators; which of
        them are used is determined by the user's options.

        Each generator should take at least one argument, ``inner``,
        and wrap the pieces of text generated by this.

        Always yield 2-tuples: (code, text). If "code" is 1, the text
        is part of the original tokensource being highlighted, if it's
        0, the text is some piece of wrapping. This makes it possible to
        use several different wrappers that process the original source
        linewise, e.g. line number generators.
        """
        source = self._format_lines(tokensource)

        # As a special case, we wrap line numbers before line highlighting
        # so the line numbers get wrapped in the highlighting tag.
        if not self.nowrap and self.linenos == 2:
            source = self._wrap_inlinelinenos(source)

        if self.hl_lines:
            source = self._highlight_lines(source)

        if not self.nowrap:
            if self.lineanchors:
                source = self._wrap_lineanchors(source)
            if self.linespans:
                source = self._wrap_linespans(source)
            # vvv customization of bepasty start:
            if self.lineparagraphs:
                source = self._wrap_lineparagraphs(source)
            # ^^^ customization of bepasty end.
            source = self.wrap(source)
            if self.linenos == 1:
                source = self._wrap_tablelinenos(source)
            source = self._wrap_div(source)
            if self.full:
                source = self._wrap_full(source, outfile)

        for t, piece in source:
            outfile.write(piece)
